static const char* const rcsid = "$Id$";

/*                               -*- Mode: C -*-
 * oprs-com.c --
 *
 * Copyright (C) 1994-2005 Francois Felix Ingrand
 *
 * Author          : Francois Felix Ingrand
 * Created On      : Wed Apr 20 16:31:25 1994
 * Last Modified By: Francois Felix Ingrand
 * Last Modified On: Mon Jun 11 16:55:42 2001
 * Update Count    : 352
 * Status          : OK
 *
 *
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>

#include <listPack.h>
#include <hashPack.h>


#include "macro.h"

#include "macro-pub.h"

#include "portLib.h"
#include "csLib.h"
#ifndef VXWORKS
#include "h2initGlob.h"
#endif
#include "h2errorLib.h"
#include "errnoLib.h"

#include "opaque-pub.h"
#include "macro-pub.h"
#include "oprs-type-pub.h"
#include "pu-mk-term_f.h"
#include "pu-parse-tl_f.h"

#include "oprs-type_f-pub.h"
#include "lisp-list_f-pub.h"
#include "fact-goal_f-pub.h"
#include "oprs_f-pub.h"


#include "user-decl-oprs-com.h"
#include "user-decl-oprs-com_f.h"
#include "oprs-com.h"
#include "oprs-com_f.h"

#include "socket_server.h"

#define RS_NAME "RS"
#define RS_REPLY_NAME "RS_REPLY"
#define RS_MSGQ_SIZE 20

#define COPY_LIST(list) copy_list((list),NOT_NODES)
#define FREE_LIST(list) free_list((list),NOT_NODES)
#define REQUEST_Q(mod) ((mod)->request_q)

typedef struct mod Mod;

struct mod {		/* Module structure */
     Symbol name;		/* its name (a symbol) */
     char *rq_name;		/* its request queue name (a string) */
     CLIENT_ID request_q;
     Client_Init client_init_func; /* Its init function (for CSLIB only) */
     PBoolean initialized;
};

typedef struct {		/* Request type */
     Symbol name;		/* Name (a symbol) */
     int cmd;			/* Server command number */
     Encode_Func_Proto encode_func; /* Encode and decode functions */
     Decode_Func_Proto decode_func;
     int encode_size;		/* Encode and decode size */
     int decode_size;
} Rqst_Type;

typedef struct {		/* An Active Request */
     int rqid;			/* its id (mine) */
     Rqst_Type *type;		/* its type */
     Mod *mod;			/* For CSLIB, ist module */
     int csLib_rqid;		/* its cslib id */
     PBoolean ir_received;	/* has it received an intermediate reply yet */
} Rqst_Instance;


/* Various hash tables */

#define MOD_HASH_SIZE 8		/* has to be a power of two */
static int mod_hashtable_size = MOD_HASH_SIZE;
static int mod_hashtable_mod = MOD_HASH_SIZE - 1;
HASH mod_hash;

#define RQST_TYPE_HASH_SIZE 128		/* has to be a power of two */
static int rqst_type_hashtable_size = RQST_TYPE_HASH_SIZE;
static int rqst_type_hashtable_mod = RQST_TYPE_HASH_SIZE -1;
HASH rqst_type_hash;

OPRS_LIST rqst_instance_list;

static PBoolean oprs_com_verbose = FALSE;

/* Hash functions. */
static int mod_hash_a_string(PString name)
{
     return (hash_a_string(name) & mod_hashtable_mod);
}
static PBoolean mod_match(PString s1, Mod *mod)
{
     return ( s1 == mod->name) ;	/* These are atoms, so they can be pointer compared */
}


static int rqst_type_hash_a_string(PString name )
{
     return (hash_a_string(name) & rqst_type_hashtable_mod);
}
static PBoolean rqst_type_match(PString s1, Rqst_Type *rqst_type)
{
     return ( s1 == rqst_type->name) ;	/* These are atoms, so they can be pointer compared */
}


/* The global reply message queue or client id. */
CLIENT_ID rsId;

static void end_com_to_module(OPRS_NODE ignore, Mod *mod)
{
     if (mod->request_q)		/* if valid request_q, shutdown it. */
	  csClientEnd(mod->request_q);
}

void end_oprs_com()
{
     /* Init the logging socket server */
     end_server();

     /* close the request queue */
     csMboxEnd ();
     for_all_hashtable(mod_hash, NULL, end_com_to_module); /* for all known modules */
}

static Pred_Func_Rec *ir_pred, *fr_pred, *failure_pred;
static Pred_Func_Rec *data_func;

void init_oprs_com(char *name)
{
#ifndef VXWORKS
     if (h2initGlob(0) == ERROR) {
	  fprintf(stderr, "oprs-com:h2initGlob error");
	  exit(1);
     }
#endif

     /* Init the logging socket server */
     start_server(3140);

     /* Init the communication system */
     if (csMboxInit (name, 0, CLIENT_NMAX_RQST_ID * RS_MAX_FR_DATA_LENGTH) == ERROR) {
	  h2perror("oprs-com:init_oprs_com:csMboxInit");
	  exit(1);
     }
     ir_pred = declare_pred_from_symbol(declare_atom("IR"));
     failure_pred = declare_pred_from_symbol(declare_atom("MOD-FAILURE-DATA"));
     fr_pred = declare_pred_from_symbol(declare_atom("FR"));

     data_func = declare_pred_from_symbol(declare_atom("DATA"));


     mod_hash = make_hashtable(mod_hashtable_size, mod_hash_a_string, mod_match);
     rqst_type_hash = make_hashtable(rqst_type_hashtable_size, rqst_type_hash_a_string, rqst_type_match);
     rqst_instance_list = make_list();
}


int declare_oprs_module(char *mod_name, char *rq_name ,Client_Init client_init_func)
{
     Mod *mod = MAKE_OBJECT(Mod);

     mod->name = declare_atom(mod_name);	/* Declare the module name symbol */
     NEW_STRING(rq_name, mod->rq_name);
     mod->request_q = NULL;
     mod->client_init_func = client_init_func;
     mod->initialized = FALSE;

     if ((get_from_hashtable(mod_hash,mod->name)) != NULL)
	  fprintf(stderr, "oprs-com:redeclaring the module:%s.\n", mod->name);

     add_to_hashtable(mod_hash, mod, mod->name); /* redeclaring it anyway... */

     return 0;
}

static PBoolean init_oprs_mod_com(Mod *mod, PBoolean force)
{
     if (!mod->initialized) {
	  if ((mod->client_init_func) && ((*(mod->client_init_func))(&(mod->request_q)) == ERROR)) {
	       h2perror("oprs-com:init_oprs_mod_com:csClientInit");
	  } else
	       mod->initialized = TRUE;
     } else if (force) {	/* Was already initialized, and we force it. */
	  if (mod->request_q)
	       csClientEnd(mod->request_q);
	  if ((mod->client_init_func) && ((*(mod->client_init_func))(&(mod->request_q)) == ERROR)) {
	       h2perror("oprs-com:init_oprs_mod_com:re csClientInit");
	       mod->initialized = FALSE;
	  } else
	       mod->initialized = TRUE;
     } else
	  fprintf(stderr, "oprs-com:init_oprs_mod_com: module: %s, already initialized.\n", mod->name);
     return mod->initialized;
}

static PBoolean init_oprs_mod_name_com(char *mod_name, PBoolean force)
{
     Mod *mod;

     if ((mod = (Mod *)get_from_hashtable(mod_hash,mod_name)) == NULL) {
	  fprintf(stderr, "oprs-com:init_oprs_mod_com:undeclared module: %s.\n", mod_name);
	  return FALSE;
     } else
	  return init_oprs_mod_com(mod,force);
}

Term *init_com_to_module(TermList terms)
/* This is the action which sends the request. */
{
     Term *t1;

     t1 = (Term *)get_list_pos(terms,1);
     if (t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM in init_com_to_module.\n");
	  return build_nil();
     } else {
	  if (init_oprs_mod_name_com(t1->u.id, FALSE))
	       return build_t();
	  else
	       return build_nil();
     }
}

Term *init_com_force_to_module(TermList terms)
/* This is the action which sends the request. */
{
     Term *t1;

     t1 = (Term *)get_list_pos(terms,1);
     if (t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM in init_com_to_module.\n");
	  return build_nil();
     } else {
	  if (init_oprs_mod_name_com(t1->u.id, TRUE))
	       return build_t();
	  else
	       return build_nil();
     }
}

void init_rqst_type(char *name, int cmd, Encode_Func_Proto encode, int encode_size,
	       Decode_Func_Proto decode, int decode_size)
{
     Rqst_Type *rqst_type = MAKE_OBJECT(Rqst_Type);

     rqst_type->name = declare_atom(name);	/* Declare the rqst type as a symbol */
     rqst_type->cmd = cmd;
     rqst_type->encode_func = encode;
     rqst_type->decode_func = decode;
     rqst_type->encode_size = encode_size;
     rqst_type->decode_size = decode_size;


     if ((get_from_hashtable(rqst_type_hash, rqst_type->name)) != NULL)
	  fprintf(stderr, "oprs-com:redeclaring the rqst type:%s.\n", rqst_type->name);

     add_to_hashtable(rqst_type_hash, rqst_type, rqst_type->name);
}

/* Function to post the intermediate reply fact. */
static void post_ir(char *from, Rqst_Instance *rqst_instance, int ir_letter)
{
     TermList terms;

     terms = make_list();
     add_to_tail(terms, build_id(from));
     add_to_tail(terms, build_id(rqst_instance->type->name));
     add_to_tail(terms, build_integer(rqst_instance->rqid));
     add_to_tail(terms, build_integer(ir_letter));
     add_fact(make_fact_from_expr(build_expr_pfr_terms(ir_pred,
						       terms))
	      ,current_oprs);
}

/* Function to post the final reply fact. */
static void post_fr(char *from, Rqst_Instance *rqst_instance, Term *term, int report)
{
     TermList terms2;
     char strerr[64];
     const char *error = h2getErrMsg(report, strerr, 64);
     char *res;

     if ((res = strchr(error,'_')) && ((res = strchr(res+1,'_'))) && *(res+1))
	  error = res+1;

     terms2 = make_list();
     add_to_tail(terms2, build_id(from));
     add_to_tail(terms2, build_id(rqst_instance->type->name));
     add_to_tail(terms2, build_integer(rqst_instance->rqid));
     add_to_tail(terms2, build_string(error));
     add_to_tail(terms2, term);

     add_fact(make_fact_from_expr(build_expr_pfr_terms(fr_pred,
						       terms2)),
	      current_oprs);
}

static void check_and_get_cs_reply(OPRS_LIST rqid_to_del, Rqst_Instance *rqst_instance)
{
     int status;
     char *fr_letter = (char *)malloc(rqst_instance->type->decode_size);
     int ir_letter;
     int report = OK;
     char strerr[64];

     status = csClientReplyRcv (REQUEST_Q(rqst_instance->mod), rqst_instance->csLib_rqid, NO_BLOCK, (char *)&ir_letter,
				sizeof(int), NULL, fr_letter, rqst_instance->type->decode_size, NULL);

     switch (status) {
     case WAITING_FINAL_REPLY:
	  if (!rqst_instance->ir_received) {
	       post_ir(rqst_instance->mod->name, rqst_instance, ir_letter);
	       rqst_instance->ir_received = TRUE;
	  }
	  break;
     case ERROR:
	  report = errnoGet();
	  if (!H2_MODULE_ERR_FLAG(report)) {
	       fprintf(stderr, "oprs-com: Error %s (id %d, csLib %d): %s.\n",
		       rqst_instance->type->name, rqst_instance->rqid, rqst_instance->csLib_rqid, h2getErrMsg(report, strerr, 64));

	       add_to_head(rqid_to_del,rqst_instance);
	       fprintf(stderr,"oprs-com:  rcv  %s %d: %s\n", rqst_instance->type->name, rqst_instance->rqid, h2getErrMsg(report, strerr, 64));
	       if (oprs_com_verbose)
		    printf("oprs-com:  rcv  %s %d: %s\n", rqst_instance->type->name, rqst_instance->rqid, h2getErrMsg(report, strerr, 64));
	       break;
	  } else {
	       fprintf(stderr, "oprs-com:  rcv  %s %d: %s\n", rqst_instance->type->name, rqst_instance->rqid, h2getErrMsg(report, strerr, 64));
	       printf("oprs-com:  rcv  %s %d: %s\n", rqst_instance->type->name, rqst_instance->rqid, h2getErrMsg(report, strerr, 64));
	  }
	  server_send("[FR %d \"%s\"]", rqst_instance->rqid, h2getErrMsg(report, strerr, 64));
     case FINAL_REPLY_OK: {
	       Term *term;

	       if (rqst_instance->type->decode_func) {
		    term = (*(rqst_instance->type->decode_func))(NULL,fr_letter,1);
	       } else {
		    term = PUMakeTermNil();
	       }
	       post_fr(rqst_instance->mod->name, rqst_instance, term, report);
	       add_to_head(rqid_to_del,rqst_instance);
	       if (status == FINAL_REPLY_OK)
		    fprintf(stderr,"oprs-com:  rcv  %s %d\n", rqst_instance->type->name, rqst_instance->rqid);
               server_send("[FR %d \"OK\"]", rqst_instance->rqid);
	       if (oprs_com_verbose && (status == FINAL_REPLY_OK)
		   /* &&  ! (strstr("IRC_SEND IRC_RECV IRC_BROADCAST CI_SEND CI_RECV", rqst_instance->type->name)) */
		      )
		    printf("oprs-com:  rcv  %s %d\n", rqst_instance->type->name, rqst_instance->rqid);

	  }
	  break;
     case WAITING_INTERMED_REPLY:
	  break;
/*      case FREE: */
     case INTERMED_REPLY_TIMEOUT:
     case FINAL_REPLY_TIMEOUT:
     default:
	  fprintf(stderr, "oprs-com:check_and_get_cs_reply: csClient_Reply_Rcv: Error in status %d.\n", status);
     }
     free(fr_letter);
}

PBoolean get_reply_and_check_com(TermList ignore)
{
     Rqst_Instance *rqst_instance;
     OPRS_LIST to_del = make_list();

     static int tick_count = 0;
     struct timespec cur_date;
     if (clock_gettime(CLOCK_REALTIME, &cur_date) == 0) {
          server_send("[TICK %d %ld%06ld]", tick_count, cur_date.tv_sec, cur_date.tv_nsec / 1000);
     }
     ++tick_count;

     loop_through_list(rqst_instance_list, rqst_instance,  Rqst_Instance *) {
	  check_and_get_cs_reply(to_del, rqst_instance);
     }

     loop_through_list(to_del, rqst_instance,  Rqst_Instance *) {
	  delete_list_node(rqst_instance_list,rqst_instance);
	  FREE_OBJECT(rqst_instance);
     }

     FREE_LIST(to_del);

     return FALSE;		/* Never return TRUE, unless we want to stop. */
}

Term *ef_strerror(TermList terms)
{
     const char *res;
     int error;
     char strerr[64];

     if (! PUGetOprsParameters(terms, 1, INTEGER, &error)) {
	  fprintf(stderr, "oprs-com:ef_srtrerror:GetOprsParameters:unable to parse request.\n");
     }

     res = h2getErrMsg(error, strerr, 64);

     return build_string(res);
}

Term *send_request_to_module(TermList terms)
/* This is the action which sends the request.
 * the format of the request is: (... MOD-NAME TYPE-NAME (DATA ...)).*/
{
     static int rqid = 1;
     Rqst_Type *rqst_type;
     Mod *mod;
     int encode_res;
     Term * res;
     PBoolean param_res;
     char *mod_name;
     char *type_name;

     if (! (param_res = PUGetOprsParameters(terms, 2, ATOM, &mod_name, ATOM, &type_name))) {
	  fprintf(stderr, "oprs-com:send_request_to_module:GetOprsParameters:unable to parse request.\n");
	  res = build_integer(-1);
     } else if ((rqst_type = (Rqst_Type *)get_from_hashtable(rqst_type_hash, type_name)) == NULL) {
	  fprintf(stderr, "oprs-com:send_request_to_module:could not find request type %s.\n", type_name);
	  res = build_integer(-1);
     } else if ((mod = (Mod *)get_from_hashtable(mod_hash, mod_name)) == NULL) {
	  fprintf(stderr, "oprs-com:send_request_to_module:could not find module %s.\n", mod_name);
	  res = build_integer(-1);
     } else if (!mod->initialized && !init_oprs_mod_com(mod,FALSE)) {
	  fprintf(stderr, "oprs-com:send_request_to_module: module %s not initialized.\n", mod_name);
	  res = build_integer(-1);
     } else {
	  Rqst_Instance *rqst_instance;
	  char *rq_letter;
	  Expression *term;

	  if (rqst_type->encode_func) {

	       if (! PUGetOprsParametersSpecArg(terms, 3, EXPRESSION, &term)) {
		    fprintf(stderr, "oprs-com:send_request_to_module:PUGetOprsParametersSpecArg:unable to parse request.\n");
		    return build_integer(-1);
	       }

	       rq_letter = (char *)malloc(rqst_type->encode_size);

	       encode_res = (*(rqst_type->encode_func))(
		    NULL,
		    term,		/* Term */
		    (void *)(rq_letter), /* Structure pointer */
		    1		/* Array 1 */
		    );
	       if (! encode_res) {
		    fprintf(stderr, "oprs-com:send_request_to_module:problem while encoding data\n\
\tfor request %s for module %s.\n", type_name, mod_name);
		    free(rq_letter);
		    return build_integer(-1);
	       }
	  } else {
	       rq_letter = NULL;
	  }

	  rqst_instance = MAKE_OBJECT(Rqst_Instance);
	  rqst_instance->rqid = rqid;
	  rqst_instance->type = rqst_type;
	  rqst_instance->mod = mod;
	  rqst_instance->ir_received = FALSE;

	  if ( csClientRqstSend (REQUEST_Q(mod), rqst_type->cmd, rq_letter,
				 rqst_type->encode_size,
				 NULL, TRUE, 0, 0, &(rqst_instance->csLib_rqid)) == ERROR) {
	       h2perror("oprs-com:send_request_to_module:csClientRqstSend");
	       /* May be we should reset the initialized bit... */
	       res = build_integer(-1);
	  } else {
	       fprintf(stderr,"oprs-com: *send %s %d\n", type_name, rqid);
               server_send("[RQ %s %d]", type_name, rqid);
	       if (oprs_com_verbose)
		    printf("oprs-com: *send %s %d\n", type_name, rqid);

	       add_to_tail(rqst_instance_list, rqst_instance);
	       res = build_integer(rqid);
	       rqid++;
	  }
	  if (rq_letter) {
	       free(rq_letter);
	  }
     }
     return res;
}

Term *set_oprs_com_verbose(TermList terms)
/* This is the action which sends the request. */
{
     Term *t1;

     t1 = (Term *)get_list_pos(terms,1);
     if (t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM in set_oprs_com_verbose.\n");
	  return build_nil();
     } else {
	  oprs_com_verbose = ! (t1->u.id == nil_sym);
	  return build_t();
     }
}

/* ---------------------------------------------------------------------- */

/*
   Usefull functions to be call in "init_rqst_type" when a request has
   no input (null_encode) or no output (null_decode) parameter.
*/
Encode_Func_Proto null_encode = NULL;

Decode_Func_Proto null_decode = NULL;

#ifdef IGNORE
int abort_encode(void *data, TermList terms)
{
     int *actid = (int *) data;
     int id;

     if (! PUGetOprsParameters(terms, 1,
			      INTEGER, &id)) {
	  fprintf(stderr, "RS: abort_encode: pb in PUGetOprsParameters\n");
	  return -1;
     }

     *actid = id;

     return sizeof(int);
}
#endif


int collect_module(OPRS_LIST res, Mod *mod)
{
     add_to_head(res, build_id(mod->name));
     return 1;
}

Term *ef_list_modules(TermList terms)
{
     OPRS_LIST res = make_list();

     sh_for_all_hashtable(mod_hash, res, collect_module);

     return build_l_list(transform_c_list_to_l_list(res));
}


int collect_initialized_module(OPRS_LIST res, Mod *mod)
{
     if (mod->initialized) {
	  add_to_head(res, build_id(mod->name));
	  return 1;
     } else return 0;
}

Term *ef_list_initialized_modules(TermList terms)
{
     OPRS_LIST res = make_list();

     sh_for_all_hashtable(mod_hash, res, collect_initialized_module);

     return build_l_list(transform_c_list_to_l_list(res));
}


int collect_rqst_type(OPRS_LIST res, Rqst_Type *rq)
{
     add_to_head(res, build_id(rq->name));
     return 1;
}

Term *ef_list_requests(TermList terms)
{
     OPRS_LIST res = make_list();

     sh_for_all_hashtable(rqst_type_hash, res, collect_rqst_type);

     return build_l_list(transform_c_list_to_l_list(res));
}
